//
// Copyright Amazon.com Inc. or its affiliates.
// All Rights Reserved.
//
// SPDX-License-Identifier: Apache-2.0
//

import Foundation

extension SRPSignInState {
    struct Resolver: StateMachineResolver {
        typealias StateType = SRPSignInState
        let defaultState = SRPSignInState.notStarted

        func resolve(
            oldState: SRPSignInState,
            byApplying event: StateMachineEvent
        ) -> StateResolution<SRPSignInState> {

            guard let srpSignInEvent = event as? SignInEvent else {
                return .from(oldState)
            }

            if case .throwAuthError(let authError) = srpSignInEvent.eventType {
                return errorState(authError)
            }

            if case .throwPasswordVerifierError(let authError) = srpSignInEvent.eventType {
                return errorState(authError)
            }

            switch oldState {
            case .notStarted:
                return resolveNotStarted(byApplying: srpSignInEvent)
            case .initiatingSRPA(let signInEventData):
                return resolveInitiatingSRPA(
                    byApplying: srpSignInEvent,
                    from: oldState)
            case .respondingPasswordVerifier(let srpStateData):
                return resolveRespondingVerifyPassword(
                    srpStateData: srpStateData,
                    byApplying: srpSignInEvent)
            case .signedIn, .error:
                return .from(oldState)
            case .cancelling:
                return .from(.notStarted)
            }
        }

        private func resolveNotStarted(byApplying signInEvent: SignInEvent) -> StateResolution<SRPSignInState> {
            switch signInEvent.eventType {
            case .initiateSignInWithSRP(let signInEventData, let deviceMetadata):
                guard let username = signInEventData.username, !username.isEmpty else {
                    let error = SignInError.inputValidation(
                        field: AuthPluginErrorConstants.signInUsernameError.field
                    )
                    return errorState(error)
                }
                let authFlowType: AuthFlowType
                switch signInEventData.signInMethod {
                case .apiBased(let flowType):
                    authFlowType = flowType
                default:
                    let error = SignInError.configuration(message: "Invalid auth flow type during SRP Sign In")
                    return errorState(error)
                }

                // Assuming password could be nil
                let password = signInEventData.password ?? ""
                let action = InitiateAuthSRP(
                    username: username,
                    password: password,
                    authFlowType: authFlowType,
                    deviceMetadata: deviceMetadata,
                    clientMetadata: signInEventData.clientMetadata)
                return StateResolution(
                    newState: SRPSignInState.initiatingSRPA(signInEventData),
                    actions: [action]
                )
            default:
                return .from(.notStarted)
            }
        }

        private func resolveInitiatingSRPA(
            byApplying signInEvent: SignInEvent,
            from oldState: SRPSignInState)
        -> StateResolution<SRPSignInState> {
            switch signInEvent.eventType {
            case .respondPasswordVerifier(let srpStateData, let authResponse, let clientMetadata):
                let action = VerifyPasswordSRP(
                    stateData: srpStateData,
                    authResponse: authResponse,
                    clientMetadata: clientMetadata)
                return StateResolution(
                    newState: SRPSignInState.respondingPasswordVerifier(srpStateData),
                    actions: [action]
                )
            case .cancelSRPSignIn:
                return .from(.cancelling)
            default:
                return .from(oldState)
            }
        }

        private func resolveRespondingVerifyPassword(
            srpStateData: SRPStateData,
            byApplying signInEvent: SignInEvent)
        -> StateResolution<SRPSignInState> {
            switch signInEvent.eventType {
            case .retryRespondPasswordVerifier(let srpStateData, let authResponse, let clientMetadata):
                let action = VerifyPasswordSRP(
                    stateData: srpStateData,
                    authResponse: authResponse,
                    clientMetadata: clientMetadata)
                return StateResolution(
                    newState: SRPSignInState.respondingPasswordVerifier(srpStateData),
                    actions: [action]
                )
            case .finalizeSignIn(let signedInData):
                return .init(newState: .signedIn(signedInData),
                             actions: [SignInComplete(signedInData: signedInData)])
            case .cancelSRPSignIn:
                return .from(.cancelling)
            default:
                return .from(.respondingPasswordVerifier(srpStateData))
            }
        }

        private func errorState(_ error: SignInError)
        -> StateResolution<SRPSignInState> {
            return StateResolution(
                newState: SRPSignInState.error(error)
            )
        }

    }
}
